import numpy as np


# Matrix for converting axial coordinates to pixel coordinates
axial_to_pixel_mat = np.array([[np.sqrt(3), np.sqrt(3) / 2], [0, 3 / 2.]])

# Matrix for converting pixel coordinates to axial coordinates
pixel_to_axial_mat = np.linalg.inv(axial_to_pixel_mat)


# These are the vectors for moving from any hex to one of its neighbors.
SE = np.array((1, 0, -1))
SW = np.array((0, 1, -1))
W = np.array((-1, 1, 0))
NW = np.array((-1, 0, 1))
NE = np.array((0, -1, 1))
E = np.array((1, -1, 0))
ALL_DIRECTIONS = np.array([NW, NE, E, SE, SW, W, ])


def get_cube_distance(hex_start, hex_end):
    """
    Computes the smallest number of hexes between hex_start and hex_end, on the hex lattice.
    :param hex_start: Starting hex...
    :param hex_end: Ending hex...
    :return: Smallest number of hexes between `hex_start` and `hex_end`, on hex lattice.
    """
    return np.sum(np.abs(hex_start - hex_end) / 2)


# Selection Functions ######

def get_neighbor(hex, direction):
    """
    Simply returns the neighbor, in the direction specified, of the hexagon.
    :param hex: Cube coordinates of the hexagon.
    :param direction: A direction from the DIR class.
    :return: The location of the neighbor in cube coordinates.
    """
    return hex + direction


def get_ring(center, radius):
    """
    Retrieves the locations of all the hexes exactly a certain distance from a hexagon.
    :param center: The location of the hexagon to get the ring of.
    :param radius: The distance from `center` of the hexes we want.
    :return: An array of locations of the hexes that are exactly `radius` units away from `center`.
    """
    if radius < 0:
        return []
    if radius == 0:
        return [center]

    rad_hex = np.zeros((6 * radius, 3))
    count = 0
    for i in range(0, 6):
        for k in range(0, radius):
            rad_hex[count] = ALL_DIRECTIONS[i - 1] * (radius - k) + ALL_DIRECTIONS[i] * (k)
            count += 1

    return np.squeeze(rad_hex) + center


def get_disk(center, radius):
    """
    Retrieves the locations of all the hexes within certain distance from a hexagon.
    :param center: The location of the hexagon to get the neighbors of.
    :param radius: The distance from `center` of the hexes we want.
    :return: An array of locations of the hexes that are within `radius` units away from `center`.
    """
    hex_set = []
    for x in range(-radius, radius + 1):
        for y in range(max(-radius, -(x + radius)), min(radius + 1, -x + radius + 1)):
            z = -x - y
            hex_set.append(np.array([x, y, z]))
    return np.array(hex_set) + center


def get_spiral(center, radius_start=1, radius_end=2):
    """
    Retrieves all hexes that are `radius` hexes away from the `center`.
    :param center: The location of the center hex.
    :param radius_start: The distance from center. We want all hexes greater than or equal to this distance.
    :param radius_end: The distance from center. We want all hexes within this distance from `center`.
    :return: An array of locations of the hexes that are within `radius` hexes away from `center`.
    """
    hex_area = get_ring(center, 0)
    for i in range(radius_start, radius_end + 1):
        hex_area = np.append(hex_area, get_ring(center, i), axis=0)
    return hex_area


def get_hex_line(hex_start, hex_end):
    """
    Get hexes on line from hex_start to hex_end.
    :param hex_start: The hex where the line starts.
    :param hex_end: The hex where the line ends.
    :return: A set of hexes along a straight line from hex_start to hex_end.
    """
    hex_distance = get_cube_distance(hex_start, hex_end)
    if hex_distance < 1:
        return np.array([hex_start])

    # Set up linear system to compute linearly interpolated cube points
    bottom_row = np.array([i / hex_distance for i in np.arange(hex_distance)])
    x = np.vstack((1 - bottom_row, bottom_row))
    A = np.vstack((hex_start, hex_end)).T

    # linearly interpolate from a to b in n steps
    interpolated_points = A.dot(x)
    interpolated_points = np.vstack((interpolated_points.T, hex_end))
    return np.array(cube_round(interpolated_points))


# Conversion Functions ######

def cube_to_axial(cube):
    """
    Convert cube to axial coordinates.
    :param cube: A coordinate in cube form. nx3
    :return: `cube` in axial form.
    """
    return np.vstack((cube[:, 0], cube[:, 2])).T


def axial_to_cube(axial):
    """
    Convert axial to cube coordinates.
    :param axial: A coordinate in axial form.
    :return: `axial` in cube form.
    """
    x = axial[:, 0]
    z = axial[:, 1]
    y = -x - z
    cube_coords = np.vstack((x, y, z)).T
    return cube_coords


def axial_to_pixel(axial, radius):
    """
    Converts the location of a hex in axial form to pixel coordinates.
    :param axial: The location of a hex in axial form. nx3
    :param radius: Radius of all hexagons.
    :return: `axial` in pixel coordinates.
    """
    pos = radius * axial_to_pixel_mat.dot(axial.T)
    return pos.T


def cube_to_pixel(cube, radius):
    """
    Converts the location of a hex in cube form to pixel coordinates.
    :param cube: The location of a hex in cube form. nx3
    :param radius: Radius of all hexagons.
    :return: `cube` in pixel coordinates.
    """
    in_axial_form = cube_to_axial(cube)
    return axial_to_pixel(in_axial_form, radius)


def pixel_to_cube(pixel, radius):
    """
    Converts the location of a hex in pixel coordinates to cube form.
    :param pixel: The location of a hex in pixel coordinates. nx2
    :param radius: Radius of all hexagons.
    :return: `pixel` in cube coordinates.
    """
    axial = pixel_to_axial_mat.dot(pixel.T) / radius
    return cube_round(axial_to_cube(axial.T))


def pixel_to_axial(pixel, radius):
    """
    Converts the location of a hex in pixel coordinates to axial form.
    :param pixel: The location of a hex in pixel coordinates. nx2
    :param radius: Radius of all hexagons.
    :return: `pixel` in axial coordinates.
    """
    cube = pixel_to_cube(pixel, radius)
    return cube_to_axial(cube)


def cube_round(cubes):
    """
    Rounds a location in cube coordinates to the center of the nearest hex.
    :param cubes: Locations in cube form. nx3
    :return: The location of the center of the nearest hex in cube coordinates.
    """
    rounded = np.zeros((cubes.shape[0], 3))
    rounded_cubes = np.round(cubes)
    for i, cube in enumerate(rounded_cubes):
        (rx, ry, rz) = cube
        xdiff, ydiff, zdiff = np.abs(cube-cubes[i])
        if xdiff > ydiff and xdiff > zdiff:
            rx = -ry - rz
        elif ydiff > zdiff:
            ry = -rx - rz
        else:
            rz = -rx - ry
        rounded[i] = (rx, ry, rz)
    return rounded


def axial_round(axial):
    """
    Rounds a location in axial coordinates to the center of the nearest hex.
    :param axial: A location in axial form. nx2
    :return: The location of the center of the nearest hex in axial coordinates.
    """
    return cube_to_axial(cube_round(axial_to_cube(axial)))



import numpy as np
import hexy as hx

class HexTile(object):
    """
    Base Hex class. Doesn't do anything. Ideally, you want to store instances of
    a subclass of this tile in a HexMap object.
    """

    def __init__(self, axial_coordinates, radius, tile_id):
        super(HexTile, self).__init__()
        self.axial_coordinates = np.array([axial_coordinates])
        self.cube_coordinates = hx.axial_to_cube(self.axial_coordinates)
        self.position = hx.axial_to_pixel(self.axial_coordinates, radius)
        self.radius = radius
        self.tile_id = tile_id





"""
This file contains routines for storing the properties of hexes

The keys are just the axial coordinates.

Key Rules:
- must be written 'a,b' without quotes
- no spaces
- a and b can be negative, so it is okay if the key is written as
  '-a,b' or 'a,-b' or '-a,-b'
"""
import numpy as np

from errors import IncorrectCoordinatesError, HexExistsError, MismatchError
from hexy import *

# The bases of the axial coordinate system
bases_mat = cube_to_axial(np.array([SE, E], dtype=int))


def make_key_from_coordinates(indexes):
    """
    Converts indexes to string for hashing
    :param indexes: the indexes of a hex. nx2, n=number of index pairs
    :return: key for hashing based on index.
    """
    return [str(int(index[0])) + ',' + str(int(index[1])) for index in indexes]


def solve_for_indexes(hexes):
    """
    We want to solve for the coefficients in the linear combos.
    :param hexes: The hexes whose indexes we want to solve for.
                  nx2, n=number of hexes
    :return: indexes of `hexes`
    """
    if hexes.shape[1] != 2:
        raise IncorrectCoordinatesError("Must be axial coordinates!")
    return np.linalg.solve(bases_mat, hexes.T).T


class HexMap(dict):
    def __init__(self):
        super(HexMap, self).__init__()

    def __setitem__(self, coordinates, hex_objects):
        """
        Assigns hex objects as values to coordinates as keys. The number of coordinates and hex objects
        should be equal.
        :param coordinates: Locations of hex objects.
        :param hex_objects: the hex objects themselves.
        :return: None
        """
        if len(coordinates) != len(hex_objects):
            raise MismatchError("Number of coordinates does not match number of hex objects.")

        keys = make_key_from_coordinates(coordinates)
        for key, hex in zip(keys, hex_objects):
            if key in self.keys():
                raise HexExistsError("key " + key + " already exists.")

            super(HexMap, self).__setitem__(key, hex)

    def setitem_direct(self, key, value):
        if key in self.keys():
            raise HexExistsError("key " + key + " already exists.")

        super(HexMap, self).__setitem__(key, value)

    def overwrite_entries(self, coordinates, hex):
        keys = make_key_from_coordinates(coordinates)
        for key in keys:
            super(HexMap, self).__setitem__(key, hex)

    def __delitem__(self, coordinates):
        if len(coordinates.shape) == 1:
            coordinates = np.array([coordinates])
        keys = make_key_from_coordinates(coordinates)
        for key in keys:
            if key in self.keys():
                super(HexMap, self).__delitem__(key)

    def __getitem__(self, coordinates):
        """
        Retrieves hexes stores at `coordinates`
        :param coordinate: the locations used as keys for hexes. You can pass more than one coordinate
        :return: list of hexes mapped to using `coordinates`
        """
        if len(coordinates.shape) == 1:
            coordinates = np.array([coordinates])
        keys = make_key_from_coordinates(coordinates)
        hexes = []
        for key in keys:
            if key in self.keys():
                hexes.append(super(HexMap, self).__getitem__(key))
        return hexes



import numpy as np
import hexy as hx


hm = hx.HexMap()
radius = 10
coords = hx.get_spiral(np.array((0, 0, 0)), 1, 10)


def test_axial_to_cube_conversion():
    axial_coords = hx.cube_to_axial(coords)
    cube_coords = hx.axial_to_cube(axial_coords)

    assert np.array_equal(coords, cube_coords)


def test_axial_to_pixel_conversion():
    axial_coords = hx.cube_to_axial(coords)
    pixel_coords = hx.axial_to_pixel(axial_coords, radius)
    pixel_to_axial_coords = hx.pixel_to_axial(pixel_coords, radius)

    assert np.array_equal(axial_coords, pixel_to_axial_coords)


def test_cube_to_pixel_conversion():
    axial_coords = hx.cube_to_axial(coords)
    cube_coords = hx.axial_to_cube(axial_coords)
    pixel_coords = hx.cube_to_pixel(cube_coords, radius)
    pixel_to_cube_coords = hx.pixel_to_cube(pixel_coords, radius)
    pixel_to_cube_to_axial_coords = hx.cube_to_axial(pixel_to_cube_coords)

    assert np.array_equal(cube_coords, pixel_to_cube_coords)


def test_the_converted_coords_and_dataset_coords_retrieve_the_same_data():
    axial_coords = hx.cube_to_axial(coords)
    cube_coords = hx.axial_to_cube(axial_coords)
    pixel_coords = hx.cube_to_pixel(cube_coords, radius)
    pixel_to_cube_coords = hx.pixel_to_cube(pixel_coords, radius)
    pixel_to_cube_to_axial_coords = hx.cube_to_axial(pixel_to_cube_coords)

    # check that we can correctly retrieve hexes after conversions
    hm[axial_coords] = coords
    retrieved = hm[pixel_to_cube_to_axial_coords]

    assert np.array_equal(retrieved, coords)




from distutils.core import setup
setup(
  name = 'Hexy',         
  packages = ['hexy'],   
  version = '1.4',      
  license='MIT',        
  description = 'A library that makes working with a hexagonal lattice easier.',   
  author = 'Norbu Tsering',                   
  author_email = 'norbu.tsering.cs@gmail.com',      
  url = 'https://github.com/redft/hexy',   
  download_url = 'https://github.com/RedFT/Hexy/archive/1.4.tar.gz',    
  keywords = ['hexy', 'coordinate', 'hexagon', 'hexagonal'],
  install_requires = ["numpy == 1.15.0"],
  extras_require ={            
      'examples': [
        "atomicwrites (==1.1.5)",
        "attrs (==18.1.0)",
        "funcsigs (==1.0.2)",
        "more-itertools (==4.3.0)",
        "pathlib2 (==2.3.2)",
        "pluggy (==0.7.1)",
        "py (==1.5.4)",
        "pygame (==1.9.4)",
        "pytest (==3.7.0)",
        "scandir (==1.7)",
        "six (==1.11.0)",
        ]
      },
  classifiers=[
    'Development Status :: 5 - Production/Stable',      
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Libraries :: Python Modules',
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 2',      
  ],
)



class HexExistsError(Exception):
    def __init__(self, message):
        super(HexExistsError, self).__init__(message)


class IncorrectCoordinatesError(Exception):
    def __init__(self, message):
        super(IncorrectCoordinatesError, self).__init__(message)


class MismatchError(Exception):
    def __init__(self, message):
        super(MismatchError, self).__init__(message)
